import asyncio
import json
import logging
import re
from pathlib import Path
from typing import Any

import aiofiles
import aiohttp
from bs4 import BeautifulSoup
from bs4.element import Tag
from rich.console import Console
from rich.logging import RichHandler

from cda_dl.download_options import DownloadOptions
from cda_dl.download_state import DownloadState
from cda_dl.error import (
    GeoBlockedError,
    HTTPError,
    LoginRequiredError,
    ParserError,
    ResolutionError,
)
from cda_dl.ui import RichUI
from cda_dl.utils import (
    decrypt_url,
    get_request,
    get_safe_title,
    get_video_match,
    post_request,
)

logging.basicConfig(
    level=logging.INFO,
    format="%(message)s",
    handlers=[RichHandler(show_time=False)],
)
LOGGER = logging.getLogger(__name__)


class Video:
    video_id: str
    resolutions: dict[str, str]
    video_soup: BeautifulSoup
    video_info: Any
    resolution: str
    file: str
    video_stream: aiohttp.ClientResponse
    remaining_size: int
    title: str
    filepath: Path
    partial_filepath: Path
    resume_point: int

    def __init__(
        self, url: str, session: aiohttp.ClientSession, ui: RichUI
    ) -> None:
        self.url = url
        self.session = session
        self.ui = ui
        self.headers = {
            "Content-Type": "application/json",
            "X-Requested-With": "XMLHttpRequest",
        }

    async def download_video(
        self, download_options: DownloadOptions, download_state: DownloadState
    ) -> None:
        LOGGER.level = (
            logging.WARNING if download_options.quiet else logging.INFO
        )
        try:
            await self.pre_initialize(download_options)
            if self.filepath.exists() and not download_options.overwrite:
                LOGGER.info(
                    f"Plik '{self.title}.mp4' już istnieje. Pomijam ..."
                )
                download_state.skipped += 1
                return
            await self.initialize(download_options)
        except (
            LoginRequiredError,
            GeoBlockedError,
            ResolutionError,
            ParserError,
            HTTPError,
        ) as e:
            if isinstance(e, HTTPError) and e.status_code == 429:
                LOGGER.warning("Zbyt dużo zapytań. Usypiam wątek na 10 min.")
                await asyncio.sleep(60 * 10)
                await self.download_video(download_options, download_state)
            else:
                LOGGER.warning(e)
                download_state.failed += 1
        else:
            self.make_directory(download_options)
            await self.stream_file(download_state)

    async def pre_initialize(self, download_options: DownloadOptions) -> None:
        """Initialize members required to get Video info."""
        self.video_soup = await self.get_video_soup()
        self.title = self.get_video_title()
        self.filepath = self.get_filepath(download_options)

    async def initialize(self, download_options: DownloadOptions) -> None:
        """Initialize members required to download the Video."""
        self.video_id = self.get_videoid()
        self.check_geolocation()
        self.partial_filepath = self.get_partial_filepath()
        self.check_premium()
        self.video_info = await self.get_video_info()
        self.resolutions = self.get_resolutions()
        self.resolution = self.get_adjusted_resolution(download_options)
        self.raise_invalid_res()
        cda_res = self.resolutions[self.resolution]
        resp = await post_request(
            self.url,
            self.session,
            {
                "id": 3,
                "jsonrpc": "2.0",
                "method": "videoGetLink",
                "params": [
                    self.video_id,
                    cda_res,
                    self.video_info["ts"],
                    self.video_info["hash2"],
                    {},
                ],
            },
            self.headers,
        )
        data = await resp.json()
        self.file = data["result"]["resp"]
        self.resume_point = self.get_resume_point()
        self.video_stream = await self.get_video_stream()
        self.remaining_size = self.get_remaining_size()

    def get_videoid(self) -> str:
        """Get videoid from Video url."""
        match = get_video_match(self.url)
        assert match
        return match.group(1)

    async def get_video_soup(self) -> BeautifulSoup:
        response = await get_request(self.url, self.session, self.headers)
        text = await response.text()
        return BeautifulSoup(text, "html.parser")

    def get_video_title(self) -> str:
        title_tag = self.video_soup.find("h1")
        if not isinstance(title_tag, Tag):
            raise ParserError(
                "Error podczas parsowania 'video title' dla"
                f" {self.url} Pomijam ..."
            )
        title = title_tag.text.strip("\n")
        return get_safe_title(title)

    def get_filepath(self, download_options: DownloadOptions) -> Path:
        return Path(download_options.directory, f"{self.title}.mp4")

    def get_partial_filepath(self) -> Path:
        return self.filepath.parent / f"{self.filepath.name}.part"

    def check_premium(self) -> None:
        if (
            "Ten film jest dostępny dla użytkowników premium"
            in self.video_soup.text
        ):
            raise LoginRequiredError(
                f"{self.title} jest dostępny tylko dla użytkowników premium."
                " Pomijam ..."
            )

    def check_geolocation(self) -> None:
        if re.search(
            r"niedostępn[ey] w(?:&nbsp;|\s+)Twoim kraju\s*",
            self.video_soup.text,
        ):
            raise GeoBlockedError(
                f"{self.url} jest niedostępny w Twoim kraju. Pomijam ..."
            )

    async def get_video_info(self) -> Any:
        """Get Video info from the url."""
        media_player = self.video_soup.find(
            "div", {"id": f"mediaplayer{self.video_id}"}
        )
        if not isinstance(media_player, Tag):
            raise ParserError(
                f"Error podczas parsowania 'media player' dla {self.title}."
                " Pomijam ..."
            )
        player_data = json.loads(media_player.attrs["player_data"])
        return player_data["video"]

    def get_resolutions(self) -> dict[str, str]:
        """Get available Video resolutions at the url."""
        return self.video_info["qualities"]  # type: ignore

    async def list_resolutions(self) -> None:
        self.video_id = self.get_videoid()
        self.video_soup = await self.get_video_soup()
        self.video_info = await self.get_video_info()
        resolutions = self.get_resolutions()
        console = Console()
        console.print(f"Dostępne rozdzielczości dla {self.url}")
        for res in resolutions:
            console.print(res)

    async def check_resolution(
        self, download_options: DownloadOptions
    ) -> None:
        self.video_id = self.get_videoid()
        self.video_soup = await self.get_video_soup()
        self.video_info = await self.get_video_info()
        self.resolutions = self.get_resolutions()
        self.resolution = download_options.resolution
        self.raise_invalid_res()

    def get_best_resolution(self) -> str:
        """Get best Video resolution available at the url."""
        numeric_resolutions = []
        for k in self.resolutions.keys():
            # Skip non-numeric resolution keys like 'aut'
            if k.endswith('p') and k[:-1].isdigit():
                numeric_resolutions.append(int(k[:-1]))
        
        if not numeric_resolutions:
            # If no numeric resolutions found, return the first available resolution
            return list(self.resolutions.keys())[0]
            
        return f"{max(numeric_resolutions)}p"

    def is_valid_resolution(self) -> bool:
        return self.resolution in self.resolutions

    def get_adjusted_resolution(
        self, download_options: DownloadOptions
    ) -> str:
        return (
            self.get_best_resolution()
            if download_options.resolution == "najlepsza"
            else download_options.resolution
        )

    def raise_invalid_res(self) -> None:
        """Raise ResolutionError if resolution is invalid."""
        if not self.is_valid_resolution():
            raise ResolutionError(
                f"{self.resolution} rozdzielczość nie jest dostępna dla"
                f" {self.url}"
            )

    def get_file(self) -> str:
        """Get decrypted link to the file download."""
        return decrypt_url(self.video_info["file"])

    def get_resume_point(self) -> int:
        return (
            self.partial_filepath.stat().st_size
            if self.partial_filepath.exists()
            else 0
        )

    async def get_video_stream(self) -> aiohttp.ClientResponse:
        range_num = f"bytes={self.resume_point}-"
        self.headers["Range"] = range_num
        video_stream = await get_request(self.file, self.session, self.headers)
        return video_stream

    def get_remaining_size(self) -> int:
        """Get remaining Video size in KiB."""
        return int(self.video_stream.headers.get("content-length", 0))

    def make_directory(self, download_options: DownloadOptions) -> None:
        download_options.directory.mkdir(parents=True, exist_ok=True)

    async def stream_file(self, download_state: DownloadState) -> None:
        block_size = 1024
        desc = f"{self.title}.mp4 [{self.resolution}]"
        self.filepath.unlink(missing_ok=True)
        assert self.ui.progbar_video
        task_id = self.ui.progbar_video.add_task(
            "download",
            filename=desc,
            total=self.resume_point + self.remaining_size,
            completed=self.resume_point,
        )
        async with aiofiles.open(self.partial_filepath, "ab") as f:
            async for chunk in self.video_stream.content.iter_chunked(
                block_size * block_size
            ):
                await f.write(chunk)
                self.ui.progbar_video.update(task_id, advance=len(chunk))
        self.partial_filepath.rename(self.filepath)
        self.ui.progbar_video.remove_task(task_id)
        download_state.completed += 1



import json
import logging
import os
import sys
from typing import Any

import pytest

sys.path.append(os.path.dirname(os.path.dirname(os.path.abspath(__file__))))
from cda_dl.downloader import Downloader
from cda_dl.main import parse_args
from cda_dl.utils import is_folder, is_video

directory = os.path.abspath(os.path.dirname(__file__))
VIDEO_DATA = json.load(open(os.path.join(directory, "video_data.json")))[
    "videos"
]
FOLDER_DATA = json.load(open(os.path.join(directory, "folder_data.json")))[
    "folders"
]


def test_is_video() -> None:
    for video in VIDEO_DATA:
        assert is_video(video["url"]) is True

    for folder in FOLDER_DATA:
        assert is_video(folder["url"]) is False


def test_is_folder() -> None:
    for video in VIDEO_DATA:
        assert is_folder(video["url"]) is False

    for folder in FOLDER_DATA:
        assert is_folder(folder["url"]) is True


def test_list_resolutions_and_exit_folder(caplog: Any) -> None:
    for folder in FOLDER_DATA:
        args = parse_args(["-R", folder["url"]])
        with pytest.raises(SystemExit, match=""):
            with caplog.at_level(logging.WARNING):
                Downloader(args)
            assert (
                "Opcja -R jest dostępna tylko dla filmów."
                f" {folder['url']} jest folderem!"
                in caplog.text
            )


def test_list_resolutions_and_exit_video(caplog: Any) -> None:
    for video in VIDEO_DATA:
        if not video["resolutions"]:
            continue
        args = parse_args(["-R", video["url"]])
        with pytest.raises(SystemExit, match=""):
            with caplog.at_level(logging.INFO):
                Downloader(args)
            assert (
                f"Dostępne rozdzielczości dla {video['url']}:" in caplog.text
            )


def test_list_resolutions_and_exit_unknown(caplog: Any) -> None:
    url = "https://www.google.com"
    args = parse_args(["-R", url])
    with pytest.raises(SystemExit, match=""):
        with caplog.at_level(logging.WARNING):
            Downloader(args)
        assert f"Nie rozpoznano adresu url: {url}" in caplog.text


def test_check_valid_resolution_folder(caplog: Any) -> None:
    for folder in FOLDER_DATA:
        res = "720p"
        args = parse_args(["-r", res, folder["url"]])
        with caplog.at_level(logging.ERROR):
            Downloader(args)
        assert (
            f"Opcja -r jest dostępna tylko dla filmów. {folder['url']} jest"
            " folderem!"
            in caplog.text
        )


def test_check_valid_resolution_video(caplog: Any) -> None:
    for video in VIDEO_DATA:
        if not video["resolutions"]:
            continue
        for res in video["invalid_resolutions"]:
            args = parse_args(["-r", res, video["url"]])
            with caplog.at_level(logging.INFO):
                Downloader(args)
            assert (
                f"{res} rozdzielczość nie jest dostępna dla {video['url']}"
                in caplog.text
            )


def test_check_valid_resolution_unknown(caplog: Any) -> None:
    url = "https://www.google.com"
    res = "720p"
    args = parse_args(["-r", res, url])
    with caplog.at_level(logging.ERROR):
        Downloader(args)
    assert f"Nie rozpoznano adresu url: {url}" in caplog.text


def test_set_threads_negative(caplog: Any) -> None:
    url = "https://www.cda.pl/video/9122600a"
    nthreads = "-5"
    args = parse_args(["-t", nthreads, url])
    with caplog.at_level(logging.ERROR):
        Downloader(args)
    assert (
        f"Opcja -t musi być większa od 0. Podano: {nthreads}." in caplog.text
    )



class FlagError(Exception):
    pass


class ParserError(Exception):
    pass


class GeoBlockedError(Exception):
    pass


class ResolutionError(Exception):
    pass


class LoginRequiredError(Exception):
    pass


class HTTPError(Exception):
    def __init__(self, message: str, status_code: int):
        super().__init__(message)
        self.message = message
        self.status_code = status_code

    def __str__(self) -> str:
        return self.message


class LoginError(Exception):
    pass


class CaptchaError(Exception):
    pass



import logging
import os
import sys
from typing import Any, cast

import pytest
from aiohttp import ClientSession
from rich.table import Table

sys.path.append(os.path.dirname(os.path.dirname(os.path.abspath(__file__))))
import json

from cda_dl.download_options import DownloadOptions
from cda_dl.download_state import DownloadState
from cda_dl.error import GeoBlockedError, LoginRequiredError, ResolutionError
from cda_dl.ui import RichUI
from cda_dl.video import Video

directory = os.path.abspath(os.path.dirname(__file__))
VIDEO_DATA = json.load(open(os.path.join(directory, "video_data.json")))[
    "videos"
]


def test_get_videoid() -> None:
    for video in VIDEO_DATA:
        v = Video(video["url"], cast(ClientSession, None), cast(RichUI, None))
        assert v.get_videoid() == video["videoid"]


@pytest.mark.location
@pytest.mark.asyncio
async def test_premium_video() -> None:
    url = "https://www.cda.pl/video/63289011/vfilm"
    async with ClientSession() as session:
        v = Video(url, session, cast(RichUI, None))
        v.video_id = v.get_videoid()
        v.video_soup = await v.get_video_soup()
        v.title = v.get_video_title()
        with pytest.raises(
            LoginRequiredError,
            match=(
                f"{v.title} jest dostępny tylko dla użytkowników premium."
                " Pomijam ..."
            ),
        ):
            v.check_premium()


@pytest.mark.location
@pytest.mark.asyncio
async def test_geoblocked() -> None:
    url = "https://www.cda.pl/video/124097194d/vfilm"
    async with ClientSession() as session:
        v = Video(url, session, cast(RichUI, None))
        v.video_id = v.get_videoid()
        v.video_soup = await v.get_video_soup()
        with pytest.raises(
            GeoBlockedError,
            match=f"{v.url} jest niedostępny w Twoim kraju. Pomijam ...",
        ):
            v.check_geolocation()


@pytest.mark.asyncio
async def test_get_resolutions() -> None:
    for video in VIDEO_DATA:
        if not video["resolutions"]:
            continue
        async with ClientSession() as session:
            v = Video(video["url"], session, cast(RichUI, None))
            v.video_id = v.get_videoid()
            v.video_soup = await v.get_video_soup()
            v.video_info = await v.get_video_info()
            assert v.get_resolutions() == video["resolutions"]


@pytest.mark.asyncio
async def test_get_adjusted_resolution() -> None:
    for video in VIDEO_DATA:
        if not video["adjusted_resolution"]:
            continue
        async with ClientSession() as session:
            download_options = DownloadOptions()
            v = Video(video["url"], session, cast(RichUI, None))
            v.video_id = v.get_videoid()
            v.video_soup = await v.get_video_soup()
            v.video_info = await v.get_video_info()
            v.resolutions = v.get_resolutions()
            assert (
                v.get_adjusted_resolution(download_options)
                == video["adjusted_resolution"]
            )


@pytest.mark.asyncio
async def test_raise_invalid_res() -> None:
    for video in VIDEO_DATA:
        if not video["invalid_resolutions"]:
            continue
        async with ClientSession() as session:
            v = Video(video["url"], session, cast(RichUI, None))
            v.video_id = v.get_videoid()
            v.video_soup = await v.get_video_soup()
            v.video_info = await v.get_video_info()
            v.resolutions = v.get_resolutions()
            for res in video["invalid_resolutions"]:
                v.resolution = res
                with pytest.raises(
                    ResolutionError,
                    match=(
                        f"{v.resolution} rozdzielczość nie jest dostępna dla"
                        f" {v.url}"
                    ),
                ):
                    v.raise_invalid_res()


@pytest.mark.asyncio
async def test_get_video_title() -> None:
    for video in VIDEO_DATA:
        async with ClientSession() as session:
            v = Video(video["url"], session, cast(RichUI, None))
            v.video_id = v.get_videoid()
            v.video_soup = await v.get_video_soup()
            assert v.get_video_title() == video["title"]


@pytest.mark.asyncio
async def test_download_video_overwrite() -> None:
    url = "https://www.cda.pl/video/7779552a9"
    ui = RichUI(Table())
    ui.set_progress_bar_video("")
    download_options = DownloadOptions(resolution="480p", overwrite=True)
    download_state = DownloadState()
    async with ClientSession() as session:
        v = Video(url, session, ui)
        await v.download_video(download_options, download_state)
        s = os.stat(v.filepath)
        assert s.st_size == v.remaining_size


@pytest.mark.asyncio
async def test_download_video_no_overwrite(caplog: Any) -> None:
    url = "https://www.cda.pl/video/7779552a9"
    ui = RichUI(Table())
    ui.set_progress_bar_video("")
    download_options = DownloadOptions(resolution="480p")
    download_state = DownloadState()
    async with ClientSession() as session:
        v = Video(url, session, ui)
        with caplog.at_level(logging.INFO):
            await v.download_video(download_options, download_state)
        assert f"Plik '{v.title}.mp4' już istnieje. Pomijam ..." in caplog.text



import asyncio
from pathlib import Path


class DownloadOptions:
    semaphore: asyncio.Semaphore

    def __init__(
        self,
        directory: Path = Path("."),
        resolution: str = "najlepsza",
        overwrite: bool = False,
        nthreads: int = 3,
        quiet: bool = False,
    ) -> None:
        self.directory = directory
        self.resolution = resolution
        self.overwrite = overwrite
        self.nthreads = nthreads
        self.quiet = quiet



import argparse
import os
import sys
from typing import Sequence

sys.path.append(os.path.dirname(os.path.dirname(os.path.abspath(__file__))))

from cda_dl.downloader import Downloader
from cda_dl.version import __version__


class CustomHelpFormatter(argparse.HelpFormatter):
    def __init__(self, prog: str):
        super().__init__(prog, max_help_position=40, width=80)

    def _format_action_invocation(self, action: argparse.Action) -> str:
        if not action.option_strings or action.nargs == 0:
            return super()._format_action_invocation(action)
        default = self._get_default_metavar_for_optional(action)
        args_string = self._format_args(action, default)
        return ", ".join(action.option_strings) + " " + args_string


def parse_args(argv: Sequence[str] | None) -> argparse.Namespace:
    def fmt(prog: str) -> CustomHelpFormatter:
        return CustomHelpFormatter(prog)

    parser = argparse.ArgumentParser(
        prog="cda-dl",
        usage="%(prog)s [OPCJE] URL [URL...]",
        description="Downloader do filmów i folderów z cda.pl",
        add_help=False,
        formatter_class=fmt,
    )
    parser.add_argument(
        "-h", "--help", action="help", help="Wyświetl tę pomoc i wyjdź"
    )
    parser.add_argument(
        "--version",
        action="version",
        version=f"{parser.prog} {__version__}",
        help="Wyświetl wersję programu",
    )
    parser.add_argument(
        "-q",
        "--quiet",
        action="store_true",
        help="Wyświetlaj tylko błędy i ostrzeżenia",
    )
    parser.add_argument(
        "-l",
        "--login",
        metavar="USER",
        dest="login",
        type=str,
        help="Zaloguj się do konta",
    )
    parser.add_argument(
        "-d",
        "--directory",
        metavar="PATH",
        dest="directory",
        type=str,
        default=".",
        help="Ustaw docelowy katalog (domyślnie '%(default)s')",
    )
    parser.add_argument(
        "-R",
        "--resolutions",
        dest="list_resolutions",
        action="store_true",
        help="Wyświetl dostępne rozdzielczości (dla filmu)",
    )
    parser.add_argument(
        "-r",
        "--resolution",
        metavar="RES",
        dest="resolution",
        type=str,
        default="najlepsza",
        help="Pobierz film w podanej rozdzielczości (domyślnie '%(default)s')",
    )
    parser.add_argument(
        "-o",
        "--overwrite",
        action="store_true",
        help="Nadpisz pliki, jeśli istnieją",
    )
    parser.add_argument(
        "-t",
        "--threads",
        metavar="N",
        dest="nthreads",
        type=int,
        default=3,
        help="Ustaw liczbę wątków (domyślnie %(default)s)",
    )
    parser.add_argument(
        "urls",
        metavar="URL",
        type=str,
        nargs="+",
        help="URL(y) do filmu(ów)/folder(ów) do pobrania",
    )
    return parser.parse_args(argv)


def main(argv: Sequence[str] | None = None) -> int:
    args = parse_args(argv)
    Downloader(args)
    return 0


if __name__ == "__main__":
    raise SystemExit(main())


# TODO: add support for list of http(s) proxies?



import argparse
import asyncio
import logging
import sys
from getpass import getpass
from os import path
from pathlib import Path

import aiohttp
from rich.console import Console
from rich.live import Live
from rich.logging import RichHandler
from rich.table import Table

from cda_dl.download_options import DownloadOptions
from cda_dl.download_state import DownloadState
from cda_dl.error import (
    CaptchaError,
    FlagError,
    HTTPError,
    LoginError,
    ParserError,
    ResolutionError,
)
from cda_dl.folder import Folder
from cda_dl.ui import RichUI
from cda_dl.utils import clear, get_random_agent, is_folder, is_video
from cda_dl.video import Video

logging.basicConfig(
    level=logging.INFO,
    format="%(message)s",
    handlers=[RichHandler(show_time=False)],
)
LOGGER = logging.getLogger(__name__)


class Downloader:
    urls: list[str]
    login: str | None
    password: str | None
    list_resolutions: bool
    download_options: DownloadOptions
    download_state: DownloadState
    ui: RichUI
    video_urls: list[str]
    folder_urls: list[str]

    def __init__(self, args: argparse.Namespace) -> None:
        self.urls = [url.strip() for url in args.urls]
        self.login, self.password = args.login, None
        if self.login is not None:
            self.password = getpass(f"Podaj hasło dla {self.login}: ")
        self.list_resolutions = args.list_resolutions
        self.download_options = DownloadOptions(
            Path(
                path.abspath(path.expanduser(path.expandvars(args.directory)))
            ),
            args.resolution,
            args.overwrite,
            args.nthreads,
            args.quiet,
        )
        self.download_state = DownloadState()
        self.ui = RichUI(Table.grid(expand=True))
        asyncio.run(self.main())

    async def main(self) -> None:
        async with aiohttp.ClientSession() as session:
            try:
                if self.login is not None and self.password is not None:
                    await self.perform_login(session)
                if self.list_resolutions:
                    await self.list_resolutions_and_exit(session)
                await self.check_valid_resolution(session)
                self.set_threads()
            except (FlagError, ResolutionError, LoginError, CaptchaError) as e:
                LOGGER.error(e)
            else:
                self.video_urls, self.folder_urls = self.get_urls()
                with Live(self.ui.table, refresh_per_second=10):
                    if len(self.folder_urls) > 0:
                        await self.download_folders(session)
                    if len(self.video_urls) > 0:
                        await self.download_videos(session)
                clear()
                console = Console()
                console.print(
                    "| [green bold]Pobrane Pliki:"
                    f" {self.download_state.completed}[/] - [yellow"
                    f" bold]Pominięte Pliki: {self.download_state.skipped}[/]"
                    " - [red bold]Nieudane Pliki:"
                    f" {self.download_state.failed}[/] |\n"
                )
                console.print("Skończono pobieranie. Enjoy :)")

    async def perform_login(self, session: aiohttp.ClientSession) -> None:
        """Log in to the session object."""
        data = {"username": self.login, "password": self.password}
        headers = {"User-Agent": get_random_agent()}
        try:
            r = await session.post(
                "https://www.cda.pl/login", headers=headers, data=data
            )
            r.raise_for_status()
        except aiohttp.ClientResponseError as e:
            raise LoginError(
                f"Nie udało się zalogować [{e.status}]: {e.message}."
            )
        text = await r.text()
        if 'Zaznacz pole "Nie jestem robotem"!' in text:
            raise CaptchaError("Nie udało się zalogować z powodu captchy.")
        elif "Zły login lub hasło!" in text:
            raise LoginError(
                "Nie udało się zalogować z powodu błędnego loginu lub hasła."
            )

    async def list_resolutions_and_exit(
        self, session: aiohttp.ClientSession
    ) -> None:
        """List available resolutions for a video and exit."""
        for url in self.urls:
            if is_video(url):
                await Video(url, session, self.ui).list_resolutions()
            elif is_folder(url):
                LOGGER.warning(
                    f"Opcja -R jest dostępna tylko dla filmów. {url} jest"
                    " folderem!"
                )
            else:
                LOGGER.warning(f"Nie rozpoznano adresu url: {url}")
            sys.exit()

    def changed_resolution(self) -> bool:
        """Check if resolution was changed by the user."""
        return self.download_options.resolution != "najlepsza"

    async def check_valid_resolution(
        self, session: aiohttp.ClientSession
    ) -> None:
        """Check if the resolution provided by the user is valid."""
        for url in self.urls:
            if self.changed_resolution():
                if is_video(url):
                    await Video(url, session, self.ui).check_resolution(
                        self.download_options
                    )
                elif is_folder(url):
                    raise FlagError(
                        f"Opcja -r jest dostępna tylko dla filmów. {url} jest"
                        " folderem!"
                    )
                else:
                    raise FlagError(f"Nie rozpoznano adresu url: {url}")

    def set_threads(self) -> None:
        """Set number of threads for download."""
        if self.download_options.nthreads <= 0:
            raise FlagError(
                "Opcja -t musi być większa od 0. Podano:"
                f" {self.download_options.nthreads}."
            )
        self.download_options.semaphore = asyncio.Semaphore(
            self.download_options.nthreads
        )

    def get_urls(self) -> tuple[list[str], list[str]]:
        """Split urls into two lists: video_urls and folder_urls."""
        video_urls: list[str] = []
        folder_urls: list[str] = []
        for url in self.urls:
            if is_video(url):
                video_urls.append(url)
            elif is_folder(url):
                folder_urls.append(url)
            else:
                LOGGER.warning(f"Nie rozpoznano adresu url: {url}")
                self.download_state.failed += 1
        return video_urls, folder_urls

    async def download_folders(self, session: aiohttp.ClientSession) -> None:
        self.ui.set_progress_bar_folder("bold yellow")
        self.ui.add_row_folder("green")
        for folder_url in self.folder_urls:
            try:
                await Folder(folder_url, session, self.ui).download_folder(
                    self.download_options, self.download_state
                )
            except (ParserError, HTTPError) as e:
                LOGGER.warning(e)
                self.download_state.failed += 1

    async def download_videos(self, session: aiohttp.ClientSession) -> None:
        if self.ui.progbar_video is None:
            self.ui.set_progress_bar_video("bold blue")
            self.ui.add_row_video("green")

        async def wrapper(video_url: str) -> None:
            async with self.download_options.semaphore:
                await Video(video_url, session, self.ui).download_video(
                    self.download_options, self.download_state
                )

        tasks = [
            asyncio.create_task(wrapper(video_url))
            for video_url in self.video_urls
        ]
        await asyncio.gather(*tasks)



from rich.panel import Panel
from rich.progress import (
    BarColumn,
    DownloadColumn,
    Progress,
    SpinnerColumn,
    TaskID,
    TextColumn,
    TimeRemainingColumn,
    TransferSpeedColumn,
)
from rich.table import Table


class RichUI:
    progbar_video: Progress | None
    progbar_folder: Progress | None
    task_id: TaskID

    def __init__(self, table: Table) -> None:
        self.table = table
        self.progbar_video = None
        self.progbar_folder = None

    def set_progress_bar_video(self, color: str) -> None:
        self.progbar_video = Progress(
            SpinnerColumn(),
            TextColumn(
                f"[{color}]{{task.fields[filename]}}",
                justify="right",
            ),
            BarColumn(bar_width=None),
            "[progress.percentage]{task.percentage:>3.1f}%",
            "•",
            DownloadColumn(),
            "•",
            TransferSpeedColumn(),
            "•",
            TimeRemainingColumn(),
            transient=True,
        )

    def set_progress_bar_folder(self, color: str) -> None:
        self.progbar_folder = Progress(
            TextColumn(
                f"[{color}]{{task.fields[filename]}}",
                justify="right",
            ),
            BarColumn(bar_width=None),
            "[progress.percentage]{task.percentage:>3.1f}%",
            "{task.completed} z {task.total} Pobranych Filmów lub/i Folderów",
            transient=True,
        )

    def add_row_video(self, border: str) -> None:
        assert self.table and self.progbar_video
        self.table.add_row(
            Panel.fit(
                self.progbar_video,
                title="Filmy",
                border_style=border,
                padding=(1, 1),
            )
        )

    def add_row_folder(self, border: str) -> None:
        assert self.table and self.progbar_folder
        self.table.add_row(
            Panel.fit(
                self.progbar_folder,
                title="Foldery",
                border_style=border,
                padding=(1, 1),
            )
        )

    def add_task_folder(self, filename: str, total: int) -> None:
        assert self.progbar_folder
        self.task_id = self.progbar_folder.add_task(
            "download folder",
            filename=filename,
            total=total,
        )

    def update_task_folder(self, advance: int) -> None:
        assert self.progbar_folder
        self.progbar_folder.update(self.task_id, advance=advance)

    def remove_task_folder(self) -> None:
        assert self.progbar_folder
        self.progbar_folder.remove_task(self.task_id)



import json
import os
import sys
from asyncio import Semaphore
from typing import cast

import pytest
from aiohttp import ClientSession
from rich.table import Table

sys.path.append(os.path.dirname(os.path.dirname(os.path.abspath(__file__))))
from cda_dl.download_options import DownloadOptions
from cda_dl.download_state import DownloadState
from cda_dl.folder import Folder
from cda_dl.ui import RichUI

directory = os.path.abspath(os.path.dirname(__file__))
FOLDER_DATA = json.load(open(os.path.join(directory, "folder_data.json")))[
    "folders"
]


def test_get_adjusted_url() -> None:
    for folder in FOLDER_DATA:
        f = Folder(
            folder["url"], cast(ClientSession, None), cast(RichUI, None)
        )
        assert f.url == folder["adjusted_url"]


def test_get_next_page_url() -> None:
    for folder in FOLDER_DATA:
        f = Folder(
            folder["url"], cast(ClientSession, None), cast(RichUI, None)
        )
        assert f.get_next_page_url() == folder["next_page_url"]


@pytest.mark.asyncio
async def test_get_folder_title() -> None:
    for folder in FOLDER_DATA:
        if not folder["title"]:
            continue
        async with ClientSession() as session:
            f = Folder(folder["url"], session, cast(RichUI, None))
            f.soup = await f.get_soup()
            assert await f.get_folder_title() == folder["title"]


@pytest.mark.asyncio
async def test_get_videos_from_current_page() -> None:
    for folder in FOLDER_DATA:
        if folder["pagenvideos"] == -1:
            continue
        async with ClientSession() as session:
            f = Folder(folder["url"], session, cast(RichUI, None))
            f.soup = await f.get_soup()
            assert (
                len(await f.get_videos_from_current_page())
                == folder["pagenvideos"]
            )


@pytest.mark.asyncio
async def test_get_videos_from_folder() -> None:
    for folder in FOLDER_DATA:
        async with ClientSession() as session:
            f = Folder(folder["url"], session, cast(RichUI, None))
            assert len(await f.get_videos_from_folder()) == folder["nvideos"]


@pytest.mark.asyncio
async def test_get_subfolders() -> None:
    for folder in FOLDER_DATA:
        if folder["nsubfolders"] == -1:
            continue
        async with ClientSession() as session:
            f = Folder(folder["url"], session, cast(RichUI, None))
            f.soup = await f.get_soup()
            assert len(await f.get_subfolders()) == folder["nsubfolders"]


@pytest.mark.asyncio
async def test_download_folder() -> None:
    url = "https://www.cda.pl/ARAN_Inc-/folder/29375711"
    ui = RichUI(Table())
    ui.set_progress_bar_folder("")
    download_options = DownloadOptions(resolution="480p", overwrite=True)
    download_options.semaphore = Semaphore(download_options.nthreads)
    download_state = DownloadState()
    async with ClientSession() as session:
        f = Folder(url, session, ui)
        await f.download_folder(download_options, download_state)
        for v in f.videos:
            s = os.stat(v.filepath)
            assert s.st_size == v.remaining_size



#!/usr/bin/env python3

import os
from typing import Any

from setuptools import find_packages, setup


def read_file(fname: str) -> str:
    with open(fname, encoding="utf-8") as f:
        return f.read()


def read_version(fname: str) -> Any:
    exec(compile(read_file(fname), fname, "exec"))
    return locals()["__version__"]


directory = os.path.abspath(os.path.dirname(__file__))

VERSION = read_version(os.path.join(directory, "cda_dl", "version.py"))

DESCRIPTION = "CLI downloader do filmów i folderów z cda.pl"

LONG_DESCRIPTION = read_file(os.path.join(directory, "README.md"))

REQUIREMENTS = read_file(
    os.path.join(directory, "requirements.txt")
).splitlines()

setup(
    name="cda-dl",
    version=VERSION,
    description=DESCRIPTION,
    author="Piotr Skowroński",
    license="MIT",
    long_description=LONG_DESCRIPTION,
    long_description_content_type="text/markdown",
    packages=find_packages(),
    classifiers=[
        "Topic :: Multimedia :: Video",
        "Environment :: Console",
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
    ],
    entry_points={
        "console_scripts": [
            "cda-dl=cda_dl.main:main",
        ],
    },
    install_requires=REQUIREMENTS,
    python_requires=">=3.10",
    extras_require={
        "formatting": [
            "black",
        ],
        "linting": [
            "flake8",
            "mypy",
        ],
        "testing": [
            "pytest",
            "pytest-asyncio",
            "tox",
        ],
    },
)



from __future__ import annotations

import asyncio
from pathlib import Path

import aiohttp
from bs4 import BeautifulSoup

from cda_dl.download_options import DownloadOptions
from cda_dl.download_state import DownloadState
from cda_dl.error import HTTPError, ParserError
from cda_dl.ui import RichUI
from cda_dl.utils import get_folder_match, get_request, get_safe_title
from cda_dl.video import Video


class Folder:
    title: str
    videos: list[Video]
    folders: list[Folder]
    soup: BeautifulSoup

    def __init__(
        self, url: str, session: aiohttp.ClientSession, ui: RichUI
    ) -> None:
        self.url = url
        self.url = self.get_adjusted_url()
        self.session = session
        self.ui = ui
        self.headers = {
            "Content-Type": "application/json",
            "X-Requested-With": "XMLHttpRequest",
        }

    def get_adjusted_url(self) -> str:
        """If the url has no page specified, add /1/ at the
        end of it, indicating that we start from the page 1."""
        if not self.url.endswith("/"):
            self.url += "/"
        match = get_folder_match(self.url)
        return self.url if match and match.group(2) else self.url + "1/"

    async def download_folder(
        self, download_options: DownloadOptions, download_state: DownloadState
    ) -> None:
        """Recursively download all videos and subfolders of the folder."""
        self.soup = await self.get_soup()
        self.title = await self.get_folder_title()
        await self.make_directory(download_options)
        self.folders = await self.get_subfolders()
        self.videos = await self.get_videos_from_folder()
        assert self.ui.progbar_folder
        self.ui.add_task_folder(
            self.title, len(self.folders) + len(self.videos)
        )
        if len(self.videos) > 0:
            await self.download_videos_from_folder(
                download_options, download_state
            )
        if len(self.folders) > 0:
            await self.download_subfolders(download_options, download_state)
        self.ui.remove_task_folder()

    async def download_subfolders(
        self, download_options: DownloadOptions, download_state: DownloadState
    ) -> None:
        """Download all subfolders of the folder."""
        for folder in self.folders:
            await folder.download_folder(download_options, download_state)
            assert self.ui.progbar_folder
            self.ui.update_task_folder(1)

    async def get_soup(self) -> BeautifulSoup:
        response = await get_request(self.url, self.session, self.headers)
        text = await response.text()
        soup = BeautifulSoup(text, "html.parser")
        return soup

    async def make_directory(self, download_options: DownloadOptions) -> None:
        """Make directory for the folder."""
        download_options.directory = Path(
            download_options.directory, self.title
        )
        download_options.directory.mkdir(parents=True, exist_ok=True)

    async def get_folder_title(self) -> str:
        try:
            title_wrapper = self.soup.find_all(
                "span", class_="folder-one-line"
            )[-1]
        except IndexError:
            raise ParserError(
                f"Error podczas parsowania 'folder title' dla {self.url}"
                " Pomijam ..."
            )
        title = title_wrapper.find("a", href=True).text
        return get_safe_title(title)

    async def get_subfolders(self) -> list[Folder]:
        """Get subfolders of the folder."""
        folders_soup = self.soup.find_all(
            "a", href=True, class_="object-folder"
        )
        folders = [
            Folder(folder["href"], self.session, self.ui)
            for folder in folders_soup
            if "data-foldery_id" in folder.attrs
        ]
        return folders

    async def download_videos_from_folder(
        self, download_options: DownloadOptions, download_state: DownloadState
    ) -> None:
        """Download all videos from the folder."""
        if self.ui.progbar_video is None:
            self.ui.set_progress_bar_video("bold blue")
            self.ui.add_row_video("green")

        async def wrapper(video: Video) -> None:
            async with download_options.semaphore:
                await video.download_video(download_options, download_state)
                assert self.ui.progbar_folder
                self.ui.update_task_folder(1)

        tasks = [asyncio.create_task(wrapper(video)) for video in self.videos]
        await asyncio.gather(*tasks)

    async def get_videos_from_folder(self) -> list[Video]:
        """Get all videos from the folder."""
        all_videos: list[Video] = []
        while True:
            try:
                videos = await self.get_videos_from_current_page()
            except HTTPError:
                break
            all_videos.extend(videos)
            self.url = self.get_next_page_url()
        return all_videos

    async def get_videos_from_current_page(self) -> list[Video]:
        """Get all videos from the current page."""
        response = await get_request(self.url, self.session, self.headers)
        text = await response.text()
        page_soup = BeautifulSoup(text, "html.parser")
        videos_soup = page_soup.find_all(
            "a", href=True, class_="thumbnail-link"
        )
        videos = [
            Video(
                "https://www.cda.pl" + video["href"],
                self.session,
                self.ui,
            )
            for video in videos_soup
        ]
        return videos

    def get_next_page_url(self) -> str:
        """Get next page of the folder."""
        match = get_folder_match(self.url)
        assert match
        page_number = int(match.group(2))
        stripped_url = match.group(1)
        return stripped_url + "/" + str(page_number + 1) + "/"



import os
import random
import re
import urllib.parse
from typing import Any

import aiohttp
from tenacity import (
    retry,
    retry_if_exception_type,
    stop_after_attempt,
    stop_after_delay,
    wait_fixed,
)

from cda_dl.error import HTTPError


def get_video_match(url: str) -> re.Match[str] | None:
    video_regex = re.compile(
        r"""https?://(?:(?:www|ebd)\.)?cda\.pl/
        (?:video|[0-9]+x[0-9]+)/([0-9a-z]+)""",
        re.VERBOSE | re.IGNORECASE,
    )
    return video_regex.match(url)


def is_video(url: str) -> bool:
    """Check if url is a cda video."""
    match = get_video_match(url)
    return match is not None


def get_folder_match(url: str) -> re.Match[str] | None:
    folder_regex1 = re.compile(
        r"""(https?://(?:www\.)?cda\.pl/(?!video)[a-z0-9_-]+/
        (?!folder/)[a-z0-9_-]+)/?(\d*)""",
        re.VERBOSE | re.IGNORECASE,
    )
    folder_regex2 = re.compile(
        r"""(https?://(?:www\.)?cda\.pl/(?!video)[a-z0-9_-]+/
        folder/\d+)/?(\d*)""",
        re.VERBOSE | re.IGNORECASE,
    )
    return folder_regex1.match(url) or folder_regex2.match(url)


def is_folder(url: str) -> bool:
    """Check if url is a cda folder."""
    match = get_folder_match(url)
    return match is not None


def get_safe_title(title: str) -> str:
    """Remove characters that are not allowed in the filename
    and convert spaces to underscores."""
    title = re.sub(r"[^\w\s-]", "", title)
    title = re.sub(r"[\s-]+", "_", title).strip("_")
    return title


# source: // https://www.cda.pl/js/player.js?t=1676342296
def decrypt_url(url: str) -> str:
    for p in ("_XDDD", "_CDA", "_ADC", "_CXD", "_QWE", "_Q5", "_IKSDE"):
        url = url.replace(p, "")
    url = urllib.parse.unquote(url)
    b = []
    for c in url:
        f = c if isinstance(c, int) else ord(c)
        b.append(chr(33 + (f + 14) % 94) if 33 <= f <= 126 else chr(f))
    a = "".join(b)
    a = a.replace(".cda.mp4", "")
    a = a.replace(".2cda.pl", ".cda.pl")
    a = a.replace(".3cda.pl", ".cda.pl")
    if "/upstream" in a:
        a = a.replace("/upstream", ".mp4/upstream")
        return "https://" + a
    return "https://" + a + ".mp4"


def get_random_agent() -> str:
    """Get random user agent."""
    USER_AGENTS = [
        (
            "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36"
            " (KHTML, like Gecko) Chrome/109.0.0.0 Safari/537.36"
        ),
        (
            "Mozilla/5.0 (Macintosh; Intel Mac OS X 10_15_7)"
            " AppleWebKit/537.36 (KHTML, like Gecko) Chrome/109.0.0.0"
            " Safari/537.36"
        ),
        (
            "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36"
            " (KHTML, like Gecko) Chrome/108.0.0.0 Safari/537.36"
        ),
        (
            "Mozilla/5.0 (Macintosh; Intel Mac OS X 10_15_7)"
            " AppleWebKit/537.36 (KHTML, like Gecko) Chrome/108.0.0.0"
            " Safari/537.36"
        ),
        (
            "Mozilla/5.0 (X11; Linux x86_64) AppleWebKit/537.36 (KHTML, like"
            " Gecko) Chrome/108.0.0.0 Safari/537.36"
        ),
        (
            "Mozilla/5.0 (Macintosh; Intel Mac OS X 10_15_7)"
            " AppleWebKit/605.1.15 (KHTML, like Gecko) Version/16.1"
            " Safari/605.1.15"
        ),
        (
            "Mozilla/5.0 (Macintosh; Intel Mac OS X 13_1) AppleWebKit/605.1.15"
            " (KHTML, like Gecko) Version/16.1 Safari/605.1.15"
        ),
    ]
    return random.choice(USER_AGENTS)


@retry(
    retry=retry_if_exception_type(HTTPError),
    wait=wait_fixed(1),
    stop=(stop_after_attempt(3) | stop_after_delay(5)),
    reraise=True,
)
async def get_request(
    url: str, session: aiohttp.ClientSession, headers: dict[str, str]
) -> aiohttp.ClientResponse:
    """Get request with random user agent."""
    headers["User-Agent"] = get_random_agent()
    try:
        response = await session.get(url, headers=headers)
        response.raise_for_status()
    except aiohttp.ClientResponseError as e:
        raise HTTPError(
            f"HTTP error [{e.status}]: {e.message}. Pomijam ...", e.status
        )
    else:
        return response


@retry(
    retry=retry_if_exception_type(HTTPError),
    wait=wait_fixed(1),
    stop=(stop_after_attempt(3) | stop_after_delay(5)),
    reraise=True,
)
async def post_request(
    url: str,
    session: aiohttp.ClientSession,
    json: dict[str, Any],
    headers: dict[str, str],
) -> aiohttp.ClientResponse:
    """Get request with random user agent."""
    headers["User-Agent"] = get_random_agent()
    try:
        response = await session.post(url, json=json, headers=headers)
        response.raise_for_status()
    except aiohttp.ClientResponseError as e:
        raise HTTPError(
            f"HTTP error [{e.status}]: {e.message}. Pomijam ...", e.status
        )
    else:
        return response


def clear() -> None:
    """Clear the terminal screen."""
    os.system("cls" if os.name == "nt" else "clear")
